#include <packs\dx11.particles\nodes\modules\Core\fxh\Core.fxh>

/* Here comes the struct definition of the particle. You have to register
   all attributes in the else-clause you want to use in the emitter code.*/
struct Particle {
	#if defined(COMPOSITESTRUCT)
  		COMPOSITESTRUCT
 	#else
		float3 position;
		float3 velocity;
		float3 force;
		float lifespan;
		float age;
	    float radius;
	    float3 origin;
	#endif
};
/* These Buffers are mandatory! They are needed for the lifecycle management
   and for the particle data itself. */
RWStructuredBuffer<Particle> ParticleBuffer : PARTICLEBUFFER;
RWStructuredBuffer<uint> EmitterCounterBuffer : EMITTERCOUNTERBUFFER;
RWStructuredBuffer<uint> AlivePointerBuffer : ALIVEPOINTERBUFFER;
RWStructuredBuffer<uint> AliveCounterBuffer : ALIVECOUNTERBUFFER;

/* These buffers feed the data to the particle. */
StructuredBuffer<float3> PositionBuffer <string uiname="Position Buffer";>;
StructuredBuffer<float3> VelocityBuffer <string uiname="Velocity Buffer";>;
StructuredBuffer<float3> ForceBuffer <string uiname="Force Buffer";>;
StructuredBuffer<float> LifespanBuffer <string uiname="Lifespan Buffer";>;
StructuredBuffer<float> RadiusBuffer <string uiname="Radius Buffer";>;
StructuredBuffer<float3> OriginBuffer <string uiname="Origin Buffer";>;

cbuffer cbuf
{
    uint EmitCount = 0;
	uint EmitterSize = 0;
	bool ForceEmission = false;
	uint OffsetEmission = 0;
	float4 color <bool color=true;String uiname="Default Color";> = { 1.0f,1.0f,1.0f,1.0f };
	float3 scale <String uiname="Default Scale";> = { 1.0f,1.0f,1.0f };	
}

struct csin
{
	uint3 DTID : SV_DispatchThreadID;
	uint3 GTID : SV_GroupThreadID;
	uint3 GID : SV_GroupID;
};

[numthreads(XTHREADS, YTHREADS, ZTHREADS)]
void CS_Emit(csin input)
{
	if(input.DTID.x >= EmitterSize) return;
	
	uint particleIndex = EMITTEROFFSET + input.DTID.x; 
	
	/* Now comes the emission part. There are two modes where particles can be
	   put:
	   1) in slots where dead particles lie ( so lifespan <= 0) 
	   2) or in a ringbuffer style where "living" could be killed (ForceEmission == 1)*/
	float currentLifespan = ParticleBuffer[particleIndex].lifespan;
	if ( currentLifespan <= 0.0f || ForceEmission){
		// Now we found a slot where a new particle can be placed
		// => the index of the slot = particleIndex 
		
		// The emittercounter can be incremented because we can emit a particle
		uint emitterCounter = EmitterCounterBuffer.IncrementCounter(); 
		if (emitterCounter >= EmitCount )return;
		
		if (ForceEmission) particleIndex = EMITTEROFFSET + ((emitterCounter + OffsetEmission) % EmitterSize) ;
		
		/* Ok.. Now we have an index in the PARTICLEBUFFER
		   and registered this index in the ALIVEPOINTERBUFFER.
		   We can go on and write all the attributes to the new particle and
		   then store the particle.*/
		
		Particle p = (Particle) 0;
		
		/* Get all the data of the incoming buffers to the particle.
		   The index to the dynamicbuffers comes from the emittercounter */
		uint size, stride;
		PositionBuffer.GetDimensions(size,stride);
		p.position = PositionBuffer[emitterCounter % size];
		
		VelocityBuffer.GetDimensions(size,stride);
		p.velocity = VelocityBuffer[emitterCounter % size];
		
		ForceBuffer.GetDimensions(size,stride);
		p.force = ForceBuffer[emitterCounter % size];

		LifespanBuffer.GetDimensions(size,stride);
		p.lifespan = LifespanBuffer[emitterCounter % size];

		RadiusBuffer.GetDimensions(size,stride);
		//p.radius = RadiusBuffer[emitterCounter % size];
		p.radius = 0.5f;
		
		OriginBuffer.GetDimensions(size,stride);
		p.origin = OriginBuffer[emitterCounter % size];	
		
		/* There are attributes that we don't want to initialize
		   with 0 by default. For example color and scale. The particlesystem
		   doesn't have scale and color attributes by default. They are added when
		   there is a color or scale modifier present. The notification about
		   their presence is handled with defines like KNOW_SCALE or KNOW_COLOR */
		#if defined(KNOW_SCALE)
            p.scale = scale;
    	#endif
		#if defined(KNOW_COLOR)
            p.color = color;
    	#endif
//      #if defined(KNOW_RADIUS)
//			
//		#endif
//		#if defined(KNOW_ORIGIN)
//			
//		#endif
		
		ParticleBuffer[particleIndex] = p;
		
		// UPDATE ALIVEPOINTERBUFFER
		uint aliveIndex = AliveCounterBuffer[0] + AliveCounterBuffer.IncrementCounter();
		AlivePointerBuffer[aliveIndex] = particleIndex;
	}
}

technique11 Emit { pass P0{SetComputeShader( CompileShader( cs_5_0, CS_Emit() ) );} }